import { IsTuple, IsUnknown } from "ts-essentials";
/**
 * as per DeepReadonly from ts-essentials but additionally accounts for the readonly/immutable types from this package.
 */
export type DeepImmutable<T> = T extends Date ? ReadonlyDate : T extends URL ? ReadonlyURL : T extends URLSearchParams ? ReadonlyURLSearchParams : T extends Map<infer K, infer V> ? ImmutableMap<DeepImmutable<K>, DeepImmutable<V>> : T extends ReadonlyMap<infer K, infer V> ? ImmutableMap<DeepImmutable<K>, DeepImmutable<V>> : T extends ImmutableMap<infer K, infer V> ? ImmutableMap<DeepImmutable<K>, DeepImmutable<V>> : T extends WeakMap<infer K, infer V> ? ReadonlyWeakMap<DeepImmutable<K>, DeepImmutable<V>> : T extends ReadonlyWeakMap<infer K, infer V> ? ReadonlyWeakMap<DeepImmutable<K>, DeepImmutable<V>> : T extends Set<infer U> ? ImmutableSet<DeepImmutable<U>> : T extends ReadonlySet<infer U> ? ImmutableSet<DeepImmutable<U>> : T extends ImmutableSet<infer U> ? ImmutableSet<DeepImmutable<U>> : T extends WeakSet<infer U> ? ReadonlyWeakSet<DeepImmutable<U>> : T extends ReadonlyWeakSet<infer U> ? ReadonlyWeakSet<DeepImmutable<U>> : T extends Promise<infer U> ? ReadonlyPromise<DeepImmutable<U>> : T extends AnyArray<infer U> ? T extends IsTuple<T> ? {
    readonly [K in keyof T]: DeepImmutable<T[K]>;
} : ImmutableArray<DeepImmutable<U>> : T extends {} ? {
    readonly [K in keyof T]: DeepImmutable<T[K]>;
} : IsUnknown<T> extends true ? unknown : Readonly<T>;
export type ReadonlyPartial<T> = Readonly<Partial<T>>;
export type ReadonlyRequired<T> = Readonly<Required<T>>;
export type ReadonlyPick<T, K extends keyof T> = Readonly<Pick<T, K>>;
export type ReadonlyRecord<K extends string | number | symbol, T> = Readonly<Record<K, T>>;
export type ImmutableShallow<T extends {}> = {
    readonly [P in keyof T & {}]: T[P];
};
export type ImmutableArray<T> = ImmutableShallow<ReadonlyArray<T>>;
export type ReadonlyPromise<T> = Readonly<Promise<T>>;
/**
 * as per AnyArray from ts-essentials but includes our PrincipledArray and ImmutableArray types as well.
 */
export type AnyArray<T> = PrincipledArray<T> | ImmutableArray<T> | ReadonlyArray<T> | Array<T>;
/**
 * Recursive machinery to implement PrincipledArray's flat method. Copied from TypeScript standard lib
 * with necessary changes to accommodate our array types.
 */
type FlatArray<Arr, Depth extends number> = {
    readonly done: Arr;
    readonly recur: Arr extends AnyArray<infer InnerArr> ? FlatArray<InnerArr, [
        -1,
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20
    ][Depth]> : Arr;
}[Depth extends -1 ? "done" : "recur"];
/**
 * Concat machinery, copied from TypeScript standard lib
 * with necessary changes to accommodate our array types.
 */
export type ConcatArray<T> = {
    readonly length: number;
    readonly [n: number]: T;
    readonly join: (separator?: string) => string;
    readonly slice: (start?: number, end?: number) => PrincipledArray<T> | ImmutableArray<T>;
};
/**
 * A principled immutable array type.
 * @see https://github.com/agiledigital/readonly-types/issues/518
 */
export type PrincipledArray<T> = ImmutableShallow<OmitStrict<ImmutableArray<T>, "map" | "filter" | "forEach" | "flatMap" | "flat" | "concat" | "slice" | "every" | "some" | "find" | "findIndex" | "reduce" | "reduceRight">> & {
    readonly map: <U, This = undefined>(callback: (value: T, index: number, array: PrincipledArray<T>) => U, thisArg?: This) => PrincipledArray<U>;
    readonly filter: <S extends T = T, This = undefined>(predicate: ((value: T, index: number, array: PrincipledArray<T>) => value is S) | ((value: T, index: number, array: PrincipledArray<T>) => boolean), thisArg?: This) => PrincipledArray<S>;
    readonly find: <S extends T = T, This = undefined>(predicate: ((value: T, index: number, obj: PrincipledArray<T>) => value is S) | ((value: T, index: number, obj: PrincipledArray<T>) => boolean), thisArg?: This) => S | undefined;
    readonly findIndex: <This = undefined>(predicate: (value: T, index: number, obj: PrincipledArray<T>) => boolean, thisArg?: This) => number;
    readonly flatMap: <U, This = undefined>(callback: (value: T, index: number, array: PrincipledArray<T>) => U | PrincipledArray<U> | ImmutableArray<U>, thisArg?: This) => PrincipledArray<U>;
    readonly flat: <A, D extends number = 1>(this: A, depth?: D) => PrincipledArray<FlatArray<A, D>>;
    readonly concat: (...items: readonly (T | ConcatArray<T>)[]) => PrincipledArray<T>;
    readonly slice: (start?: number, end?: number) => PrincipledArray<T>;
    readonly every: <This = undefined>(predicate: (value: T, index: number, array: PrincipledArray<T>) => boolean, thisArg?: This) => boolean;
    readonly some: <This = undefined>(predicate: (value: T, index: number, array: PrincipledArray<T>) => boolean, thisArg?: This) => boolean;
    readonly reduce: <U>(callback: (previousValue: U, currentValue: T, currentIndex: number, array: PrincipledArray<T>) => U, initialValue: U) => U;
    readonly reduceRight: <U>(callback: (previousValue: U, currentValue: T, currentIndex: number, array: PrincipledArray<T>) => U, initialValue: U) => U;
};
/**
 * Copies the provided immutable array and returns the result as a principled array.
 */
export declare const principledArray: <T>(immutableArray: ImmutableArray<T>) => PrincipledArray<T>;
export type ImmutableSet<T> = Readonly<ReadonlySet<T>>;
export type ImmutableMap<K, V> = Readonly<ReadonlyMap<K, V>>;
/**
 * Drop keys `K` from `T`, where `K` must exist in `T`.
 *
 * @see https://github.com/pelotom/type-zoo
 * @see https://github.com/Microsoft/TypeScript/issues/12215#issuecomment-377567046
 */
export type OmitStrict<T, K extends keyof T> = T extends any ? Pick<T, Exclude<keyof T, K>> : never;
export type ReadonlyURLSearchParams = Readonly<OmitStrict<URLSearchParams, "append" | "delete" | "set" | "sort" | "forEach">>;
export declare const readonlyURLSearchParams: (init?: readonly (readonly string[])[] | string[][] | Record<string, string> | string | URLSearchParams | ReadonlyURLSearchParams) => ReadonlyURLSearchParams;
export type ReadonlyURL = Readonly<OmitStrict<URL, "searchParams">> & {
    readonly searchParams: ReadonlyURLSearchParams;
};
export declare const readonlyURL: (url: string, base?: string | URL | ReadonlyURL) => ReadonlyURL | undefined;
export type ReadonlyDate = Readonly<OmitStrict<Date, "setTime" | "setMilliseconds" | "setUTCMilliseconds" | "setSeconds" | "setUTCSeconds" | "setMinutes" | "setUTCMinutes" | "setHours" | "setUTCHours" | "setDate" | "setUTCDate" | "setMonth" | "setUTCMonth" | "setFullYear" | "setUTCFullYear">>;
export declare const readonlyDate: (value: number | string | Date | ReadonlyDate) => ReadonlyDate;
export declare const validReadonlyDate: (value: number | string | Date | ReadonlyDate) => ReadonlyDate | undefined;
export declare const readonlyNow: () => number;
export type ReadonlyWeakSet<T extends object> = Readonly<OmitStrict<WeakSet<T>, "add" | "delete">>;
export type ReadonlyWeakMap<K extends object, V> = Readonly<OmitStrict<WeakMap<K, V>, "delete" | "set">>;
export declare const readonlySet: <T>(values: Iterable<T>) => Readonly<ReadonlySet<T>>;
export declare const readonlyWeakSet: <T extends object>(values: Iterable<T>) => Readonly<Pick<WeakSet<T>, typeof Symbol.toStringTag | "has">>;
export declare const readonlyMap: <K, V>(values: Iterable<readonly [K, V]>) => Readonly<ReadonlyMap<K, V>>;
export declare const readonlyWeakMap: <K extends object, V>(values: Iterable<readonly [K, V]>) => Readonly<Pick<WeakMap<K, V>, typeof Symbol.toStringTag | "has" | "get">>;
export type ReadonlyError = Readonly<Error>;
export type ReadonlyEvalError = Readonly<EvalError>;
export type ReadonlyRangeError = Readonly<RangeError>;
export type ReadonlyReferenceError = Readonly<ReferenceError>;
export type ReadonlySyntaxError = Readonly<SyntaxError>;
export type ReadonlyTypeError = Readonly<TypeError>;
export type ReadonlyURIError = Readonly<URIError>;
export type ReadonlyRegExp = Readonly<RegExp>;
export {};
